import threading
import sys
import time
import struct
from array import array

from pyee import EventEmitter
from .control_codes import Control, Program, Preference, OutgoingUSBMessageType, State, IncomingUSBMessageType, \
    Data, MessageType
from .exceptions import TimeoutError
from .log import log

CONNECT_TIMEOUT = 5
CONNECT_TIMEOUT_MESSAGE = 'Timed out Connecting - Please make sure robot is available and try again'
DISCONNECT_TIMEOUT_MESSAGE = 'Timed out Disconnecting'


def make_communicator(manager):

    class Communicator(EventEmitter):
        def __init__(self, name, mac_type, mac):
            super(Communicator, self).__init__()

            self.name = name
            self.mac_type = mac_type
            self.mac = mac
            self.mac_str = ':'.join(str(i) for i in mac)

            self.robot = None
            self.connection_handle = None

            self._event = None

            self.packet_handlers = {
                MessageType.Data: {
                    Data.Ultrasonic: self.handle_ultrasonic_data,
                    Data.Accelerometer: self.handle_accelerometer_data,
                    Data.Battery: lambda x: None,
                    Data.Gyroscope: lambda x: None,
                    Data.Temperature: lambda x: None,
                },

                MessageType.Program: {
                    Program.DoneWaiting: self.handle_wait_completed,
                },

                MessageType.Preference: {
                    Preference.MotorLeftFlip: self.handle_preference,
                    Preference.MotorRightFlip: self.handle_preference,
                    Preference.UltrasonicFiltering: self.handle_preference,
                },

                MessageType.State: {
                    State.Undefined: self.handle_state,
                },
            }

            manager.communicators.append(self)

        def connect(self):
            self._event = threading.Event()

            manager.serial.send(OutgoingUSBMessageType.Connect, [7, self.mac_type] + self.mac)
            if not self._event.wait(CONNECT_TIMEOUT):
                raise TimeoutError()

            time.sleep(0.3)

        def disconnect(self):
            self._event = threading.Event()

            manager.serial.pass_through(self.connection_handle, [MessageType.Control, Control.Disconnect, 0])
            if not self._event.wait(CONNECT_TIMEOUT):
                raise TimeoutError()

        def send(self, command, message):
            manager.serial.send(command, message)

        def pass_through(self, message):
            manager.serial.pass_through(self.connection_handle, message)

        def handle_ultrasonic_data(self, packet):
            valid = packet[3]
            distance = packet[4] << 8 | packet[5]

            if valid:
                self.robot.sensor_data[Data.Ultrasonic] = distance
            else:
                self.robot.sensor_data[Data.Ultrasonic] = sys.maxsize

        def handle_accelerometer_data(self, packet):
            x = struct.unpack_from('>i', array('B', packet[3:7]))[0]
            y = struct.unpack_from('>i', array('B', packet[7:11]))[0]
            z = struct.unpack_from('>i', array('B', packet[11:]))[0]

            self.robot.sensor_data[Data.Accelerometer] = {
                'x': x / 1000,
                'y': y / 1000,
                'z': z / 1000
            }

        def handle_wait_completed(self, packet):
            self.safe_set(self.robot)

        def handle_firmware_version(self, packet):
            version = ''.join(chr(i) for i in packet)

            self.robot.firmware_version = version

            self.safe_set(self.robot)

        def handle_hardware_version(self, packet):
            version = ''.join(chr(i) for i in packet)

            self.robot.hardware_version = version

            self.safe_set(self.robot)

        def handle_connected(self, packet):
            self.robot.connected = True
            self.connection_handle = packet[6]

            manager.emit('connected', self.robot)

            self.safe_set(self)

        def handle_disconnected(self, packet):
            self.robot.connected = False
            self.connection_handle = None

            manager.emit('disconnected', self.robot)

            self.safe_set(self)

        def handle_preference(self, packet):
            length = packet[2]

            if length == 1:
                self.robot.preferences.preferences[packet[1]] = packet[3]

            self.safe_set(self.robot.preferences)

        def handle_state(self, packet):
            length = packet[2]

            if length == 1:
                self.robot.state.state[packet[1]] = packet[3]

            self.safe_set(self.robot.state)

        def handle_robot_packet(self, packet):
            type = packet[0]
            subtype = packet[1]

            if type in self.packet_handlers:
                handlers = self.packet_handlers[type]

                if subtype in handlers:
                    handlers[subtype](packet)

        @staticmethod
        def handle_new_robot(length, packet):
            mac_type = packet[0]
            mac = packet[1:6]
            mac_str = ':'.join(str(i) for i in mac)
            name = ''.join(chr(i) for i in packet[7:])

            communicator = manager.get_communicator(mac=mac)

            if communicator is None:
                communicator = Communicator(name, mac_type, mac)

            robot = manager.Robot(communicator)

            manager.scanning_robots[mac_str] = robot
            manager.emit('robots_found', manager.scanning_robots)

        @staticmethod
        def handle_connected_robot(length, packet):
            mac = packet[0:5]
            handle = packet[6]

            for communicator in manager.communicators:
                if communicator.connection_handle == handle:
                    communicator.connection_handle = None

            communicator = manager.get_communicator(mac=mac)

            if communicator is not None:
                communicator.handle_connected(packet)

        @staticmethod
        def handle_disconnected_robot(length, packet):
            mac = packet[0:5]

            communicator = manager.get_communicator(mac=mac)

            if communicator is not None:
                communicator.handle_disconnected(packet)

        @staticmethod
        def conn_handle_wrapper(func):
            def wrapper(packet_length, packet):
                connection_handle = packet[1]
                message = packet[3:3+packet_length] if packet_length > 3 else None

                communicator = manager.get_communicator(connection_handle=connection_handle)

                if communicator is not None:
                    func(communicator, message)

            return wrapper

        @staticmethod
        def handle_wrapper(func):
            def wrapper(packet_length, packet):
                message = packet[2:3+packet_length] if packet_length > 3 else None
                func(packet_length, message)

            return wrapper

        @staticmethod
        def handle_packet(packet):
            packet_length = len(packet)
            code = packet[0]

            if code in Communicator.handlers:
                Communicator.handlers[code](packet_length, packet)

        @staticmethod
        def debug(length, packet):
            log(''.join(chr(i) for i in packet), 'debug')

        @staticmethod
        def safe_set(this):
            if this._event is not None and not this._event.is_set():
                this._event.set()
                this._event.clear()

    Communicator.handlers = {
        IncomingUSBMessageType.NewRobot: Communicator.handle_wrapper(Communicator.handle_new_robot),
        IncomingUSBMessageType.RobotConnected: Communicator.handle_wrapper(Communicator.handle_connected_robot),
        IncomingUSBMessageType.RobotDisconnected: Communicator.handle_wrapper(Communicator.handle_disconnected_robot),
        IncomingUSBMessageType.FirmwareVersion: Communicator.conn_handle_wrapper(Communicator.handle_firmware_version),
        IncomingUSBMessageType.HardwareVersion: Communicator.conn_handle_wrapper(Communicator.handle_hardware_version),
        IncomingUSBMessageType.RobotPacket: Communicator.conn_handle_wrapper(Communicator.handle_robot_packet),
        IncomingUSBMessageType.Debug: Communicator.debug,
    }

    return Communicator
