import threading
import sys

from locorobo.utils import compare_semver
from .control_codes import Control, WaitType, OutgoingUSBMessageType, Data, MessageType
from .exceptions import APIError

MISSING_WAIT_ERROR = 'Must call setupWait before using move with wait'
ROTATION_WAIT_ERROR = 'Move must be called with different directions and same speeds when using Rotation based wait'
DISTANCE_WAIT_ERROR = 'Move must be called with same directions and same speeds when using Distance based wait'


ENCODER_MOVE_TOGGLE_MINIMUM_FIRMWARE_VERSION = '1.4.0'
SUPPORTED_HARDWARE_VERSIONS = [
    '2.0',
    '2.1',
]

MINIMUM_FIRMWARE_VERSION = '1.2.0'


def make_robot(manager):
    class Robot:
        def __init__(self, communicator):

            self.communicator = communicator
            communicator.robot = self

            self.firmware_version = None
            self.hardware_version = None

            self.last_wait = None

            self.connected = False

            self.sensor_data = {
                Data.Ultrasonic: sys.maxsize,
                Data.Accelerometer: {
                    'x': 0,
                    'y': 0,
                    'z': 0
                }
            }

            self._event = threading.Event()

            self.name = communicator.name

            self.encoder_toggle_supported = True

        def connect(self):
            self.communicator.connect()
            self.check_versions()
            self.check_features()

        def disconnect(self):
            self.communicator.disconnect()

        def activate_motors(self):
            self.communicator.pass_through([MessageType.Control, Control.ActivateMotors, 0])

        def deactivate_motors(self):
            self.communicator.pass_through([MessageType.Control, Control.DeactivateMotors, 0])

        def move(self, lmd, rmd, lms, rms, wait=False, encoders=True):
            if wait and self.last_wait is None:
                raise APIError(MISSING_WAIT_ERROR)
            elif wait and self.last_wait is not None and self.last_wait['sensor'] == WaitType.Rotation and (lmd == rmd or rms != lms):
                raise APIError(ROTATION_WAIT_ERROR)
            elif wait and self.last_wait is not None and self.last_wait['sensor'] == WaitType.Distance and (lmd != rmd or rms != lms):
                raise APIError(ROTATION_WAIT_ERROR)

            packet = [
                MessageType.Control,
                Control.Motors,
                5,
                lmd,
                rmd,
                int(lms * 100),
                int(rms * 100),
                1 if wait else 0
            ]

            if self.encoder_toggle_supported:
                packet.append(1 if encoders else 0)
                packet[2] = 6

            self.communicator.pass_through(packet)

            if wait:
                self._event.wait()

        def setup_wait(self, sensor, value=0):
            b0 = (value & 255 << 24) >> 24
            b1 = (value & 255 << 16) >> 16
            b2 = (value & 255 << 8) >> 8
            b3 = value & 255

            packet = [
                MessageType.Control,
                Control.SetupWait,
                5,
                sensor,
                b0, b1, b2, b3
            ]

            self.communicator.pass_through(packet)

            self.last_wait = {
                'sensor': sensor,
                'value': value
            }

        def set_light(self, light, red, green, blue):
            if type(light) is bool:
                packet = [
                    MessageType.Control,
                    Control.Lights,
                    2,
                    light,
                    1 if red else 0
                ]
            else:
                packet = [
                    MessageType.Control,
                    Control.LightsRGB,
                    4,
                    light,
                    red,
                    green,
                    blue
                ]

            self.communicator.pass_through(packet)

        def sync_lights(self):
            packet = [
                MessageType.Control,
                Control.SyncLightsRGB,
                0
            ]

            self.communicator.pass_through(packet)

        def set_lights(self, red, green, blue):
            for i in range(8):
                self.set_light(i, red, green, blue)

        def play_note(self, note, duration, wait=False):
            packet = [
                MessageType.Control,
                Control.PlayNote,
                4,
                note,
                duration >> 8,
                duration & 0xff,
                1 if wait else 0
            ]

            self.communicator.pass_through(packet)

            if wait:
                self._event.wait()

        def play_song(self, song, wait=False):
            packet = [
                MessageType.Control,
                Control.PlaySong,
                2,
                song,
                1 if wait else 0
            ]

            self.communicator.pass_through(packet)

            if wait:
                self._event.wait()

        def enable_sensor(self, sensor, on):
            packet = [
                MessageType.Data,
                sensor,
                1,
                on
            ]

            self.communicator.pass_through(packet)

        def get_sensor_value(self, sensor):
            return self.sensor_data[sensor]

        def check_versions(self):
            self.get_firmware_version()
            self.get_hardware_version()

            if compare_semver(MINIMUM_FIRMWARE_VERSION, self.firmware_version) == 1:
                raise APIError('Incompatible firmware version. Please check for an update using the Communicator')
            elif self.hardware_version[0] != '2':
                raise APIError('Incompatible hardware version.')

        def check_features(self):
            if self.hardware_version == '2.0':
                self.encoder_toggle_supported = \
                    not (compare_semver(ENCODER_MOVE_TOGGLE_MINIMUM_FIRMWARE_VERSION, self.firmware_version) == 1)

        def get_firmware_version(self):
            self.communicator.send(OutgoingUSBMessageType.RequestFirmwareVersion, [self.communicator.connection_handle])

            self._event.wait()

        def get_hardware_version(self):
            self.communicator.send(OutgoingUSBMessageType.RequestHardwareVersion, [self.communicator.connection_handle])

            self._event.wait()

    return Robot
