import threading
import collections
import serial

from pyee import EventEmitter

from .control_codes import MessageLengthIndexes, OutgoingUSBMessageType
from .log import log


class AsyncSerial(EventEmitter, threading.Thread):
    def __init__(self, address):
        EventEmitter.__init__(self)
        threading.Thread.__init__(self)

        self.sp = serial.Serial(address, timeout=1)
        self.alive = True
        self.lock = threading.Lock()

    def stop(self):
        self.alive = False

    def run(self):
        while self.alive and self.sp.isOpen():
            try:
                data = self.sp.read(self.sp.inWaiting() or 1)
                data = list(data)
            except serial.SerialException as e:
                self.emit('closed', e)
                break

            if data:
                self.emit('data', data)

        self.emit('closed', None)
        self.stop()

    def write(self, data):
        if not isinstance(data, collections.Iterable):
            data = [data]

        with self.lock:
            self.sp.write(data)

    def close(self):
        with self.lock:
            self.stop()
            self.sp.close()


class Serial(EventEmitter):
    def __init__(self, address):
        super(Serial, self).__init__()

        self.asyncSerial = AsyncSerial(address)
        self.buffer = None

    def start(self):
        self.asyncSerial.on('closed', self.closed)
        self.asyncSerial.on('data', self.data)
        self.asyncSerial.on('error', self.error)

        self.asyncSerial.start()

    def stop(self):
        self.asyncSerial.stop()

    def pass_through(self, handle, packet):
        packet = [OutgoingUSBMessageType.PassThrough, handle, len(packet)] + packet
        self.write(packet)

    def send(self, command, message):
        data = [command]

        if not isinstance(data, collections.Iterable):
            message = [message]

        data += message

        self.write(data)

    def write(self, packet):
        log(['--------------- Sending ---------------'], type='output')
        log(['packet', packet], type='output')

        self.asyncSerial.write(packet)

    def data(self, packet):
        if self.buffer is not None:
            self.buffer += packet

            packet = self.buffer
            self.buffer = None

        if len(packet) < 3:
            self.buffer = packet
            return

        code = packet[0]
        message_length = packet[MessageLengthIndexes.get(code)]

        if message_length + 3 < len(packet):
            self.buffer = packet
            return

        self.buffer = None

        log(['--------------- Receiving ---------------'], type='input')
        log(['packet', packet], type='input')

        self.emit('data', packet)

        if message_length + 3 < len(packet):
            self.buffer = None

            packet_leftover = packet[3:]
            self.data(packet_leftover)

    def closed(self, err):
        self.emit('closed', err)

    def error(self, err):
        self.emit('error', err)
